import glob
import os
import shutil
import tempfile

import sltxpkg.config as sc
import sltxpkg.data.recipes
import sltxpkg.globals as sg
import sltxpkg.lithie.compile.tools as tools
import sltxpkg.util as su
import yaml
from importlib_resources import files

# TODO: in case of error pack the directory with tar and ship it to the folder
# TODO: show the log dump (like jake)


class Recipe():

    settings = {
        'name': '',
        'author': '',
        'hooks': {
            'pre': [],
            'in': [], # before retrieval
            'post': [] # before cleanup
        },
        # cleanup will be determined by global settings
        'cleanup_cmds': [],
        'args': '',
        'eargs': [],
        'tools': [],
        # to retrieve files:
        'wanted_files': [],
        'executable': '',
        'run': []
    }

    def __init__(self, recipe_path: str):
        super().__init__()
        recipe_full_path = recipe_path
        if not os.path.isfile(recipe_full_path):
            recipe_full_path = str(
                files(sltxpkg.data.recipes).joinpath(recipe_path))
        if not os.path.isfile(recipe_full_path):
            print("Recipe", recipe_full_path, "was not found. Exiting.")
            exit(1)
        print("Loading recipe:", recipe_full_path)
        y_conf = su.load_yaml(recipe_full_path)
        self.settings = {**self.settings, **y_conf}
        self.__process_tools()
        self.__sanitize_eargs() # we need them as a single string

    def __process_tools(self):
        for tool in self.settings['tools']:
            getattr(tools, 'tool_' + tool)(self)

    def __sanitize_eargs(self):
        self.settings['eargs'] = " ".join(self.settings['eargs'])

    @staticmethod
    def get_default_recipes() -> [str]:
        return [f for f in os.listdir() if f.endswith(".recipe")]

    # format
    def __f(self, t : str) -> str:
        for _ in range(sg.configuration[sg.C_FORMAT_MAX]):
            t = t.format(**self.settings, **sg.configuration, file=sg.args.file,
                            filenoext=os.path.splitext(sg.args.file)[0],
                            tmp=tempfile.gettempdir())
        return t

    def __runcmds(self, cmds : [str]):
        for cmd in cmds:
            cmd = self.__f(cmd) # expand
            print("  -",cmd)
            os.system(cmd)

    def __runhooks(self, hookid : str):
        print("> Hooks for \"" + hookid + "\"")
        self.__runcmds(self.settings['hooks'][hookid])

    def __critical_abort(self, code : int):
        print("Collecting files in working directory")
        archive = shutil.make_archive(os.path.join(os.getcwd(), 'sltx-log-' + su.get_now()), 'gztar', sg.configuration[sg.C_WORKING_DIR])
        print("  - Created: \"" + archive + "\" (" + os.path.basename(archive)+ ")")
        exit(code)

    # Run the recipe
    def run(self):
        sc.assure_dirs() # Ensure Working diSr and texmf home
        print(self.__f("> Running recipe \"{name}\" by \"{author}\"."))
        self.__runhooks('pre')

        print("> Running the compile commands")
        for cmd in self.settings['run']:
            cmd = self.__f(cmd) # expand
            print("  -",cmd)
            fback = os.system(cmd)
            if fback != 0:
                print("\033[31m  ! The command failed. Initiating critical abort...\033[m")
                self.__critical_abort(fback)
        
        self.__runhooks('in')

        our_dir = os.getcwd()

        # Get the resulting file(s)
        print("> Retrieving resulting files to \"" + our_dir + "\"")
        for wf in self.settings['wanted_files']:
            wf = self.__f(wf)
            print("  - Retrieving files for pattern \"" + wf + "\"")
            wanted = glob.glob(os.path.join(sg.configuration[sg.C_WORKING_DIR], wf))
            for f in wanted:
                print("Saving \"" + f + "\" ")
                shutil.copy(f, our_dir)

        self.__runhooks('post')
        if sg.configuration['cleanup']:
            print("> Cleaning up (configured by configuration)")
            self.__runcmds(self.settings['cleanup_cmds'])



import glob
import os
import re
import shutil
import sys
from concurrent import futures
from subprocess import PIPE, Popen  # execution

import sltxpkg.globals as sg
from sltxpkg.config import load_dependencies_config, write_to_log
from sltxpkg.globals import (C_AUTODETECT_DRIVERS, C_CLEANUP, C_CREATE_DIRS,
                             C_DRIVER_LOG, C_DRIVER_PATTERNS, C_DRIVERS,
                             C_RECURSIVE, C_TEX_HOME, C_WORKING_DIR,
                             DEFAULT_CONFIG, print_idx)
import sltxpkg.util as su

loaded = []


def detect_driver(idx: str, url: str):
    print_idx(idx, " - Autodetecting driver...")
    for key, patterns in sg.configuration[C_DRIVER_PATTERNS].items():
        for pattern in patterns:
            if re.search(pattern, url):
                return key
    print_idx(idx, " ! No driver found...")
    sys.exit(1)


def grab_files_from(idx: str, path: str, data: dict):
    print_idx(idx, " - Grabby-Grab-Grab files from \"" + path + "\"...")
    if "grab-files" not in data:
        print_idx(idx, " ! Key 'grab-files' not found. Won't grab any files")
        return False

    files = []
    for grab_pattern in data["grab-files"]:
        # maybe forbid level up?
        files.extend(glob.glob(os.path.join(
            path, grab_pattern), recursive=True))
    # extra so i can setup installer afterwards more easily
    print_idx(idx, " > Grabbing the follwing files for installation:",
              [os.path.relpath(f, path) for f in files])
    for file in files:
        shutil.copy2(file, su.get_tex_home())
    return True


def grab_dirs_from(idx: str, path: str, data: dict):
    print_idx(idx, " - Grabby-Grab-Grab dirs from \"" + path + "\"...")
    if "grab-dirs" not in data:
        print_idx(idx, " ! Key 'grab-dirs' not found. Won't grab any directories")
        return False

    dirs = []
    for grab_pattern in data["grab-dirs"]:
        # maybe forbid level up?
        dirs.extend(glob.glob(os.path.join(
            path, grab_pattern), recursive=True))

    # extra so i can setup installer afterwards more easily
    print_idx(idx, " > Grabbing the follwing dirs for installation:", dirs)
    for dir in dirs:
        # if fails rethrow :D
        dir_target = os.path.join(
            su.get_tex_home(), os.path.relpath(dir, path))
        if sys.version_info >= (3, 8, 0):
            # we have exist is ok
            shutil.copytree(dir, dir_target, dirs_exist_ok=True)
        else:
            # we use distutils
            print("Python version below 3.8, falling back with distutils!")
            import distutils.dir_util as du
            du.copy_tree(dir, dir_target)
    return True


def write_proc_to_log(idx: int, stream, mirror: bool):
    while True:
        line = stream.readline()
        if not line:
            break
        line_utf8 = line.decode('utf-8')
        write_to_log(line_utf8)
        if mirror:
            print_idx(idx, line_utf8)


def grab_stuff(idx: str, dep_name: str, target_dir: str, data: dict):
    print_idx(idx, " > Grabbing dependencies for " + dep_name)
    got_files = grab_files_from(idx, target_dir, data)
    got_dirs = grab_dirs_from(idx, target_dir, data)
    if not got_files and not got_dirs:
        print_idx(idx, " ! No grabs performed!")
        write_to_log("No grabs performed for: " + dep_name)


def get_target_dir(data: dict, dep_name: str, driver: str):
    return sg.configuration[C_DRIVERS][driver]["target-dir"].format(
        **data, **sg.configuration, dep_name=dep_name)


def recursive_dependencies(idx: str, target_dir: str, data: dict, dep_name: str):
    if 'dep' not in data:
        print_idx(idx, "No 'dep' key found for dep: " + dep_name +
                  " using this dep-name as default (" + sg.args.dep + ")")
        data['dep'] = sg.args.dep
    dep_files = glob.glob(os.path.join(
        target_dir, data['dep']), recursive=True)
    print_idx(idx, " - Found dep-config:", dep_files)

    if len(dep_files) <= 0:
        return

    new_dependencies = {}
    for dep_file in dep_files:
        new_dependencies = load_dependencies_config(dep_file, new_dependencies)

    install_dependencies(idx, new_dependencies)


def use_driver(idx: str, data: dict, dep_name: str, driver: str, url: str):
    # default no arguments
    if "args" not in data:
        data["args"] = ""
    driver_data = sg.configuration[C_DRIVERS][driver]
    command = driver_data["command"].format(
        **data, **sg.configuration, dep_name=dep_name)
    target_dir = get_target_dir(data, dep_name, driver)
    if os.path.isdir(target_dir) and driver_data["needs-delete"]:
        print_idx(idx, " - Target folder", target_dir,
                  "exists. Will be deleted as the driver needs this")
        shutil.rmtree(target_dir)
    print_idx(idx, " > Executing:", command)
    feedback = Popen(command, stdout=PIPE, stderr=PIPE, shell=True)
    return_code = feedback.wait()
    write_proc_to_log(idx, feedback.stdout, False)
    if return_code != 0:
        print_idx(idx, " - Error-Log of Driver:")
    write_proc_to_log(idx, feedback.stderr, return_code != 0)

    if(sg.configuration[C_RECURSIVE]):
        recursive_dependencies(idx, target_dir, data, dep_name)

    if return_code != 0:
        print_idx(idx, " ! Driver failed with code", feedback, "exiting.")
        sys.exit(return_code)

    grab_stuff(idx, dep_name, target_dir, data)


def install_dependency(name: str, idx: str, data: dict):
    print_idx(idx, "Loading \"" + name + "\"")
    if "url" not in data:
        print_idx(idx, " ! The dependency did not have an url-tag attached")
    url = data["url"]
    print_idx(idx, " - Loading from: \"" + url + "\"")
    if "driver" not in data:
        if not sg.configuration[C_AUTODETECT_DRIVERS]:
            print_idx(idx, " ! No driver given and autodetection disabled!")
        else:
            data["driver"] = detect_driver(idx, url)
    driver = data["driver"]
    print_idx(idx, " - Using driver: \"" + driver + "\"")

    if name in loaded:
        print_idx(idx, " > Skipping retrieval", name,
                  " as it was already loaded by another dep.")
        grab_stuff(idx, name, get_target_dir(data, name, driver), data)
        return
    loaded.append(name)

    if driver not in sg.configuration[C_DRIVERS]:
        print_idx(idx, " ! The selected driver is unknown. Loaded:",
                  sg.configuration[C_DRIVERS])
        sys.exit(2)

    use_driver(idx, data, name, driver, url)


def install_dependencies(idx: int, dep_dict: dict, first: bool = False):
    with futures.ThreadPoolExecutor(max_workers=sg.args.threads) as pool:
        runners = []
        for i, dep in enumerate(dep_dict['dependencies']):
            runners.append(pool.submit(install_dependency, dep, str(
                i) if first else str(idx) + "." + str(i), dep_dict['dependencies'][dep]))
        futures.wait(runners)
        for runner in runners:
            if runner.result() is not None:
                print(runner.result())



# Docker control
import json
import os

import docker
from sltxpkg.globals import DOCKER_URL


class DockerCtrl:
    def __init__(self):
        self.client = docker.from_env()
        self.client.ping()

    def update_img(self, profile: str):
        target = DOCKER_URL.format(**locals())
        print("Pulling image:", target, "this may take a few minutes")
        for line in self.client.api.pull(target, tag='latest', stream=True):
            # default values
            line = line.decode('utf-8')
            lines = line.split('\r\n')
            for subline in lines:
                if subline is None or subline.strip() == "":
                    continue
                d = {'status': 'unknown', 'progress': '', 'id': ''}
                d = {**d, **json.loads(subline)}
                print("   {status} {progress} {id}".format(**d))

    def run_in_container(self, profile: str, command: str):
        if profile.startswith(":"):
            target = profile[1:]
        else:
            target = DOCKER_URL.format(**locals())
        print("Launching container", target)
        run = self.client.containers.run(
            target, command=command, detach=True, remove=True, working_dir='/root/data',
            user='root', privileged=True, network_mode='bridge',
            volumes={
                os.getcwd(): {
                    'bind': '/root/data',
                    'mount': 'rw'
                }
            })
        for l in run.logs(stdout=True, stderr=True, stream=True, timestamps=True):
            print(l.decode('utf-8'), end='')
        print("Container completed.")



import yaml
import os
import sys
from sltxpkg import globals as sg, util as su
from sltxpkg.globals import DEFAULT_CONFIG, C_AUTODETECT_DRIVERS, C_TEX_HOME, C_WORKING_DIR, C_CREATE_DIRS, C_CLEANUP, C_DRIVER_LOG, C_DRIVERS, C_DRIVER_PATTERNS

def write_to_log(data: str):
    if sg.configuration[C_DRIVER_LOG].strip():
        with open(sg.configuration[C_DRIVER_LOG], 'a') as f:
            f.write(data)
            if not data.endswith('\n'):
                f.write("\n")


def load_configuration(file: str):
    y_conf = su.load_yaml(file)
    sg.configuration = {**sg.configuration, **y_conf}


def load_dependencies_config(file: str, target: dict):
    y_dep = su.load_yaml(file)
    return {**target, **y_dep}


def assure_dirs():
    target_path = su.get_tex_home()
    create = sg.configuration[C_CREATE_DIRS]
    if not os.path.isdir(target_path):
        if create:
            print("> Tex-Home-Dir", target_path, "not found. Creating...")
            os.makedirs(target_path)
        else:
            print("! Not allowed to create texhome. Exit")
            sys.exit(1)
    sg.configuration[C_TEX_HOME] = target_path  # expansion
    target_path = os.path.expanduser(sg.configuration[C_WORKING_DIR])
    if not os.path.isdir(target_path):
        if create:
            print("> Working-Dir", target_path, "not found. Creating...")
            os.makedirs(target_path)
        else:
            print("! Not allowed to create working-dir. Exit")
            sys.exit(1)
    sg.configuration[C_WORKING_DIR] = target_path  # expansion



import os  # list directory
import shutil  # clean working dir
import sys
import re

from pathlib import Path

import sltxpkg.util as su
from sltxpkg import dep, generate, globals as sg
from sltxpkg.config import (assure_dirs, load_configuration,
                            load_dependencies_config, write_to_log)
from sltxpkg.dep import install_dependencies
from sltxpkg.globals import (C_DRIVER_LOG, C_CLEANUP, C_CREATE_DIRS,
                             C_DRIVER_LOG, C_DRIVER_PATTERNS, C_DRIVERS,
                             C_TEX_HOME, C_WORKING_DIR, DEFAULT_CONFIG, C_RECURSIVE, C_USE_DOCKER)
import sltxpkg.lithie.compile.cooker as cooker

from sltxpkg.lithie import commands as lithiecmd

from concurrent import futures


def cmd_dependency():
    if sg.args.dep is None or sg.args.dep == "":
        print("You must supply a dependency 'file'.")
        exit(1)

    if os.path.isfile(DEFAULT_CONFIG):
        print("Automatically loading '{DEFAULT_CONFIG}'".format(
            **locals(), **globals()))
        load_configuration(DEFAULT_CONFIG)

    if sg.args.config is not None:
        load_configuration(sg.args.config)
    if sg.args.dep is not None:
        sg.dependencies = load_dependencies_config(
            sg.args.dep, sg.dependencies)

    assure_dirs()

    if "target" not in sg.dependencies or "dependencies" not in sg.dependencies:
        print("The dependency-file must supply a 'target' and an 'dependencies' key!")
        sys.exit(1)

    write_to_log("====Dependencies for:" + sg.dependencies["target"]+"\n")
    print()
    print("Dependencies for:", sg.dependencies["target"])
    print("Installing to:", su.get_tex_home())
    print()

    install_dependencies(0, sg.dependencies, first=True)

    # all installed
    if sg.configuration[C_CLEANUP]:
        print("> Cleaning up the working directory, as set.")
        shutil.rmtree(sg.configuration[C_WORKING_DIR])
    print("Loaded:", dep.loaded)
    if not sg.configuration[C_RECURSIVE]:
        print("Recursion was disabled.")
    print("Dependency installation for",
          sg.dependencies["target"], "completed.")


def cmd_version():
    print("This is sltx, a simple latex helper-utility")
    print("Tex-Home", su.get_tex_home())
    print("Version:", su.get_version())


def cmd_docker():
    lithiecmd.install()


def cmd_raw_compile():
    cooker.cook()

def cmd_compile():
    if(sg.configuration[C_USE_DOCKER]):
        print("Using docker to compile")
        lithiecmd.compile()
    else:
        print("Docker was disabled, using local compilation.")
        cmd_raw_compile()


def cmd_gen_gha():
    generate.generate()

def should_be_excluded(file : str):
    if sg.args.exclude_patterns is None:
        return False

    for exclude_pattern in sg.args.exclude_patterns:
        if re.match(exclude_pattern, file):
            return True
    return False

def cmd_cleanse():
    # Delete all current log files
    # TODO: make this dry. avoid specifying the log files multiple times (see Recipe)
    print("Cleaning local logs...")
    clean_pattern = 'sltx-log-*.tar.gz'
    for f in Path(".").glob(clean_pattern):
        if should_be_excluded(str(f)):
            print("File", f, "excluded.")
        else:   
            f.unlink()
    if sg.args.cleanse_all: 
        thome = su.get_tex_home()
        if os.path.isdir(thome):
            print("Cleaning sltx-texmf-tree... (" + thome + ")")
            shutil.rmtree(thome)
        else:
            print("The local sltx-texmf tree in \"" + thome + "\" was not found. Skipping...")


# cookes the recipe :D

from sltxpkg.lithie.compile.recipe import Recipe
import sltxpkg.globals as sg

def cook():
    recipe = Recipe('default-latexmk.recipe' if sg.args.recipe is None else sg.args.recipe)
    recipe.run()


import mmap
import os
from sys import platform
from time import localtime, strftime

import yaml
from importlib_resources import as_file, files

import sltxpkg.data
import sltxpkg.globals as sg


def default_texmf():
    if platform == "linux" or platform == "linux2":
        return "~/texmf"
    elif platform == "darwin":
        return "~/Library/texmf"
    elif platform == "win32":
        return "~/texmf"


def get_version():
    return files(sltxpkg.data).joinpath('version.info').read_text()


def load_yaml(file_path: str):
    with open(file_path, 'r') as yaml_file:
        # FullLoader only available for 5.1 and above:
        if float(yaml.__version__[:yaml.__version__.rfind(".")]) >= 5.1:
            return yaml.load(yaml_file, Loader=yaml.FullLoader)
        else:
            return yaml.load(yaml_file)


def file_contains(path: str, txt: str):
    with open(path, 'rb', 0) as file, \
            mmap.mmap(file.fileno(), 0, access=mmap.ACCESS_READ) as s:
        return s.find(txt.encode('utf-8')) != -1


def get_now():
    return strftime("%Y-%m-%d__%H-%M-%S", localtime())


def get_tex_home() -> str:
    return os.path.expanduser(sg.configuration[sg.C_TEX_HOME].format(
        **sg.configuration, os_default_texmf=default_texmf()))



import os

import yaml
from yaml.representer import SafeRepresenter

from sltxpkg.globals import C_TEX_HOME

# https://stackoverflow.com/a/20863889
def change_style(style, representer):
    def new_representer(dumper, data):
        scalar = representer(dumper, data)
        scalar.style = style
        return scalar
    return new_representer


class YamlBlock(str):
    pass


represent_literal_str = change_style('|', SafeRepresenter.represent_str)
yaml.add_representer(YamlBlock, represent_literal_str)
import sltxpkg.prompt as prompt

def assure_workflow_target(path: str):
    if os.path.isfile(path):
        print("A workflow-file with the given name does already exist!")
        overwrite = prompt.get_bool(default=False)
        if not overwrite:
            print("Aborting...")
            exit(1)
    basepath = os.path.dirname(path)
    if basepath is None or basepath.strip() == "":
        return

    if not os.path.isdir(basepath):
        print("The directory", basepath, "does not exist. should it be created?")
        create = prompt.get_bool(default=True)
        if not create:
            print("Aborting...")
            exit(1)
        os.makedirs(basepath)


def add_step(document: dict, name: str, uses: str = None, _with: dict = None, _run: str = None):
    step = {'name': name}
    if uses is not None:
        step['uses'] = uses
    if _with is not None:
        step['with'] = _with
    if _run is not None:
        step['run'] = _run
    document['jobs']['build']['steps'].append(step)


def step_checkout(document: dict):
    lfs = prompt.get_bool("Checkout lfs ", default=True)
    submodules = prompt.get_bool("Checkout submodules ", default=True)
    add_step(document,
             "Checkout git repository",
             uses="actions/checkout@v2",
             _with={
                 'lfs': lfs,
                 'submodules': submodules
             })


def step_setup_python(document: dict):
    add_step(document,
             "Setup Python environment for sltx",
             uses="actions/setup-python@v2",
             _with={
                 'python-version': '3.8'
             })


def step_setup_sltx(document: dict):
    print("Do you need your own config file for sltx?")
    own_config = prompt.get_bool(default=False)
    config_file = ".sltx-gh-conf.yaml"
    texmf_home = "./texmf/tex/latex/sltx"
    dep_file = "sltx-dep.yml"
    setup_lines = ""

    if own_config:
        config_file = prompt.get_file("Path to config-file")
        print("Make sure you have set {C_TEX_HOME} to {texmf_home}".format(
            **globals(), **locals()))
    else:
        setup_lines += "echo \"{C_TEX_HOME}: {texmf_home}\" > \"{config_file}\"\n".format(
            **globals(), **locals())

    dep_file = prompt.get_file("Path to dep-file [{default}]", default=dep_file)

    exec_line = "sltx --config \"{config_file}\" --dependencies \"{dep_file}\"".format(
        **locals())

    add_step(document,
             "Setup and run sltx-install",
             _run=YamlBlock("pip install sltx\n" +
                            setup_lines + exec_line + "\n")
             )


def step_compile(document: dict):
    print("Which documents do you want to have compiled? You may separate multiple ones by comma.")
    file_list = prompt.get("File(s)")
    files = file_list.split(',')

    exec_lines = "tlmgr conf texmf TEXMFHOME \"~/Library/texmf:./texmf\"\n"

    args = "-pdf -file-line-error -halt-on-error -interaction=nonstopmode"

    # TODO: fail as soon as one fails
    for file in files:
        exec_lines += "latexmk " + args + " \"" + file + "\"\n"

    add_step(document,
             "Setup Texlive and compile the documents",
             uses="xu-cheng/texlive-action/full@v1",
             _with={
                 'run': YamlBlock(exec_lines)
             })
    return files


def step_commit_and_push(document: dict, files: list):
    print("Do you want the resulting pdf(s) to be pushed?")
    do_push = prompt.get_bool(default=True)

    if not do_push:
        return

    files = [os.path.splitext(f)[0] + ".pdf" for f in files]

    print("Which documents do you want to have pushed? You may separate multiple ones by comma.")
    print("Default:", files)
    push_file_list = prompt.get("Doc(s)", default="")
    if push_file_list.strip() != "":
        files = push_file_list.split(',')

    add_line = "git add -f " + " ".join(["\"" + f + "\"" for f in files]) + "\n"

    # commit
    add_step(document,
             "Commit",
             _run=YamlBlock(
                 "git config --local user.email \"action@github.com\"\ngit config --local user.name \"GitHub Action\"\n" + 
                 add_line + "git commit -m \"Newly compiled data\"\n"))

    # push
    branch = prompt.get("Push-Branch [{default}]", default="gh-pages")
    add_step(document,
             "Push changes",
             uses="ad-m/github-push-action@master",
             _with={
                    'branch': branch,
                    'github_token': "${{ secrets.GITHUB_TOKEN }}",
                    'force': True
             })


def generate():
    document = {}

    print("We will now generate a GitHub-Action workflow")
    target_path = prompt.get(
        "Workflow-Path [{default}]", default=".github/workflows/compile.yaml")
    assure_workflow_target(target_path)

    document['name'] = prompt.get("Workflow name")
    document['on'] = {'push': {'branches': ['master']}}
    document['jobs'] = {
        'build': {
            'runs-on': 'ubuntu-latest',
            'steps': []
        }
    }

    step_checkout(document)
    step_setup_python(document)
    step_setup_sltx(document)
    files = step_compile(document)
    step_commit_and_push(document, files)

    print("Ok, I will write the file now...")
    with open(target_path, 'tw') as f:
        # We will disable this to get name on top and have no sorting
        stream = yaml.dump(document, default_flow_style=False, sort_keys=False)
        # formatting a little bit?
        f.write(stream.replace('jobs:', '\njobs:'))
    print("File written to \"" + target_path + "\". Job completed.")



import os

def valid_default(inp: str):
    return inp is not None and inp.strip() != ""


def valid_dir(inp: str):
    return valid_default(inp) and os.path.isdir(inp)


def valid_file(inp: str):
    return valid_default(inp) and os.path.isfile(inp)


def get(prompt: str, valid_input=valid_default, type=str, default=None):
    got = None
    prompt = "\033[36m" + prompt.format(**locals()) + ">\033[m "
    if default is None:
        while not valid_input(got):
            got = input(prompt)
    else:
        got = input(prompt)
    ret = type(default if not valid_input(got) else got)
    return ret


def get_bool(pre="", default=None):
    if default is None:
        prompt = "[true/false]"
    elif default:
        prompt = "[TRUE/false]"
    else:
        prompt = "[True/FALSE]"

    return get(pre + prompt, type=bool, default=default)


def get_dir(prompt: str, default=None):
    return get(prompt, valid_input=valid_dir, type=str, default=default)


def get_file(prompt: str, default=None):
    return get(prompt, valid_input=valid_file, type=str, default=default)